/*  Extenmote : NES, SNES, N64 and Gamecube to Wii remote adapter firmware
 *  Copyright (C) 2012  Raphael Assenat <raph@raphnet.net>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>
#include <math.h>
#include "wiimote.h"
#include "gcn64_protocol.h"
#include "snes.h"
#include "n64.h"
#include "gamecube.h"
#include "rlut.h"

static unsigned char classic_id[6] = { 0x00, 0x00, 0xA4, 0x20, 0x01, 0x01 };
static unsigned char cal_data[32];
static volatile char performupdate;

static void hwInit(void)
{
	/* PORTC 
	 * 
	 * 0: SNES/NES CLK 		out1
	 * 1: SNES/NES LATCH	out1
	 * 2: SNES/NES DATA		in-pu
	 * 3: N64/Gamecube IO	in
	 * 4: SDA (io)
	 * 5: SCL (io)
	 * 6: N/A
	 * 7: N/A
	 *
	 * [*] Needs an external 1.5k pullup. An internal pullup is 
	 *     not enough (too weak). It is enabled here in case the project 
	 *     is assembled for NES/SNES only (i.e. without the external pullup).
	 *     We do not want floating inputs.
	 *
	 */
	PORTC = 0x07;
	DDRC = 0xc3;

	/* PORTB
	 *
	 * 0: out0
	 * 1: out0
	 * 2: out0
	 * 3: out0
	 * 4: out0
	 * 5: out0
	 * 6: out0
	 * 7: out0
	 *
	 */
	PORTB = 0x00;
	DDRB = 0xff;

	/* PORTD
	 * 0: out0
	 * 1: out0
	 * 2: out0
	 * 3: out0
	 * 4: out1 // Tied to VCC on Multiuse PCB2 for routing reasons
	 * 5: out0
	 * 6: out0
	 * 7: out0
	 */
	PORTD = 0x10;
	DDRD = 0xff;

}

#define ANALOG_STYLE_DEFAULT	0
#define ANALOG_STYLE_N64		1
#define ANALOG_STYLE_GC			2

void pack_classic_data(classic_pad_data *src, unsigned char dst[6], int analog_style)
{
	unsigned char rx,ry,lx=0x20,ly=0x20; // down sized
	char x,y;

	if (analog_style == ANALOG_STYLE_N64)
	{
#ifdef WITH_N64	
		x = src->lx;
		y = src->ly;

		if (x>=0) {
			lx = 0x20 + rlut7to5(x);
		} else {
			lx = 0x20 - rlut7to5(-x);
		}

		if (y>=0) {
			ly = 0x20 + rlut7to5(y);
		} else {
			ly = 0x20 - rlut7to5(-y);
		}
#endif
	} else if (analog_style == ANALOG_STYLE_GC) {
#ifdef WITH_GAMECUBE
		int nx,ny;
		nx = src->lx / 5;
		ny = src->ly / 5;
		if (nx > 0x1F)
			nx = 0x1f;
		if (ny > 0x1f)
			ny = 0x1f;
		if (nx <= -0x1f) {
			nx = -0x1f;
		}
		if (ny <= -0x1f) {
			ny = -0x1f;
		}
		lx = 0x20 + nx;
		ly = 0x20 + ny;
#endif
	}
	else {
		lx = ((0x80 + src->lx) >> 2) & 0x3F;
		ly = ((0x80 + src->ly) >> 2) & 0x3F;
	}

	rx = ((0x80 + src->rx) >> 3) & 0x1F;
	ry = ((0x80 + src->ry) >> 3) & 0x1F;


//	lx = lut64[lx];
//	ly = lut64[ly];

	dst[0] = ((rx<<3) & 0xC0) | lx;
	dst[1] = ((rx<<5) & 0xC0) | ly;
	dst[2] = rx<<7 | ry;

	dst[3] = 0xff;
	dst[4] = (src->buttons >> 8) ^ 0xFF;
	dst[5] = (src->buttons) ^ 0xFF;
}

#define MODE_STANDARD			0
#define MODE_MARIOKART64		1
#define MODE_OCARINA			2
#define MODE_SSMB				3
#define MODE_SIN_AND_PUNISHMENT	4
#define MODE_OGRE_BATTLE		5
static unsigned char g_n64_mapping_mode = MODE_STANDARD;

static unsigned char g_snes_nes_mode = 0;
static unsigned char g_snes_analog_dpad = 0;

#define IS_SIMULTANEOUS(x,mask)	(((x)&(mask)) == (mask))

void dataToClassic(const gamepad_data *src, classic_pad_data *dst)
{
	dst->buttons = 0;
	dst->rx = dst->ry = 0;
	dst->lx = dst->ly = 0;

	switch (src->pad_type)
	{
		case PAD_TYPE_NONE:
			break;

		case PAD_TYPE_SNES:
			if (g_snes_nes_mode) {
				if (src->snes.buttons & SNES_BTN_Y) { dst->buttons |= CPAD_BTN_B; }
				if (src->snes.buttons & SNES_BTN_B) { dst->buttons |= CPAD_BTN_A; }
			} else {
				if (src->snes.buttons & SNES_BTN_B) { dst->buttons |= CPAD_BTN_B; }
				if (src->snes.buttons & SNES_BTN_Y) { dst->buttons |= CPAD_BTN_Y; }
				if (src->snes.buttons & SNES_BTN_A) { dst->buttons |= CPAD_BTN_A; }
				if (src->snes.buttons & SNES_BTN_X) { dst->buttons |= CPAD_BTN_X; }
			}

			if (g_snes_analog_dpad) {
				if (src->snes.buttons & SNES_BTN_DPAD_UP) { dst->ly = 100; }
				if (src->snes.buttons & SNES_BTN_DPAD_DOWN) { dst->ly = -100; }
				if (src->snes.buttons & SNES_BTN_DPAD_LEFT) { dst->lx = -100; }
				if (src->snes.buttons & SNES_BTN_DPAD_RIGHT) { dst->lx = 100; }
			} else {
				if (src->snes.buttons & SNES_BTN_DPAD_UP) { dst->buttons |= CPAD_BTN_DPAD_UP; }
				if (src->snes.buttons & SNES_BTN_DPAD_DOWN) { dst->buttons |= CPAD_BTN_DPAD_DOWN; }
				if (src->snes.buttons & SNES_BTN_DPAD_LEFT) { dst->buttons |= CPAD_BTN_DPAD_LEFT; }
				if (src->snes.buttons & SNES_BTN_DPAD_RIGHT) { dst->buttons |= CPAD_BTN_DPAD_RIGHT; }
			}

			if (src->snes.buttons & SNES_BTN_SELECT) { dst->buttons |= CPAD_BTN_MINUS; }
			if (src->snes.buttons & SNES_BTN_START) { dst->buttons |= CPAD_BTN_PLUS; }
			if (src->snes.buttons & SNES_BTN_L) { dst->buttons |= CPAD_BTN_TRIG_LEFT; }
			if (src->snes.buttons & SNES_BTN_R) { dst->buttons |= CPAD_BTN_TRIG_RIGHT; }
	
			if (IS_SIMULTANEOUS(src->snes.buttons, SNES_BTN_START|SNES_BTN_SELECT|SNES_BTN_L|SNES_BTN_R|SNES_BTN_DPAD_UP)) {
				g_snes_nes_mode = 0;
				g_snes_analog_dpad = 0;
			}
			if (IS_SIMULTANEOUS(src->snes.buttons, SNES_BTN_START|SNES_BTN_SELECT|SNES_BTN_L|SNES_BTN_R|SNES_BTN_DPAD_DOWN)) {
				g_snes_nes_mode = 1;
				g_snes_analog_dpad = 0;
			}
			if (IS_SIMULTANEOUS(src->snes.buttons, SNES_BTN_START|SNES_BTN_SELECT|SNES_BTN_L|SNES_BTN_R|SNES_BTN_DPAD_LEFT)) {
				g_snes_analog_dpad = 1;
				g_snes_nes_mode = 0;
			}
		
			break;

		case PAD_TYPE_NES:
			if (src->nes.buttons & NES_BTN_A) { dst->buttons |= CPAD_BTN_A; }
			if (src->nes.buttons & NES_BTN_B) { dst->buttons |= CPAD_BTN_B; }
			if (src->nes.buttons & NES_BTN_SELECT) { dst->buttons |= CPAD_BTN_MINUS; }
			if (src->nes.buttons & NES_BTN_START) { dst->buttons |= CPAD_BTN_PLUS; }
			if (src->nes.buttons & NES_BTN_DPAD_UP) { dst->buttons |= CPAD_BTN_DPAD_UP; }
			if (src->nes.buttons & NES_BTN_DPAD_DOWN) { dst->buttons |= CPAD_BTN_DPAD_DOWN; }
			if (src->nes.buttons & NES_BTN_DPAD_LEFT) { dst->buttons |= CPAD_BTN_DPAD_LEFT; }
			if (src->nes.buttons & NES_BTN_DPAD_RIGHT) { dst->buttons |= CPAD_BTN_DPAD_RIGHT; }
			break;

#ifdef WITH_GAMECUBE
		case PAD_TYPE_GAMECUBE:
			if (src->gc.buttons & GC_BTN_A) { dst->buttons |= CPAD_BTN_A; }
			if (src->gc.buttons & GC_BTN_B) { dst->buttons |= CPAD_BTN_B; }
			if (src->gc.buttons & GC_BTN_X) { dst->buttons |= CPAD_BTN_X; }
			if (src->gc.buttons & GC_BTN_Y) { dst->buttons |= CPAD_BTN_Y; }
			if (src->gc.buttons & GC_BTN_L) { dst->buttons |= CPAD_BTN_TRIG_LEFT; }
			if (src->gc.buttons & GC_BTN_R) { dst->buttons |= CPAD_BTN_TRIG_RIGHT; }
			if (src->gc.buttons & GC_BTN_START) { dst->buttons |= CPAD_BTN_PLUS; }
			if (src->gc.buttons & GC_BTN_Z) { dst->buttons |= CPAD_BTN_ZL|CPAD_BTN_ZR; }
			if (src->gc.buttons & GC_BTN_DPAD_UP) { dst->buttons |= CPAD_BTN_DPAD_UP; }	
			if (src->gc.buttons & GC_BTN_DPAD_DOWN) { dst->buttons |= CPAD_BTN_DPAD_DOWN; }	
			if (src->gc.buttons & GC_BTN_DPAD_LEFT) { dst->buttons |= CPAD_BTN_DPAD_LEFT; }	
			if (src->gc.buttons & GC_BTN_DPAD_RIGHT) { dst->buttons |= CPAD_BTN_DPAD_RIGHT; }	

			dst->lx = src->gc.x;
			dst->ly = src->gc.y;
			break;
#endif

#ifdef WITH_N64
		case PAD_TYPE_N64:
			if (IS_SIMULTANEOUS(src->n64.buttons, N64_BTN_L|N64_BTN_R|N64_BTN_Z|N64_BTN_DPAD_UP)) {
				g_n64_mapping_mode = MODE_STANDARD;
			} else if (IS_SIMULTANEOUS(src->n64.buttons, N64_BTN_L|N64_BTN_R|N64_BTN_Z|N64_BTN_DPAD_DOWN)) {
				g_n64_mapping_mode = MODE_MARIOKART64;
			} else if (IS_SIMULTANEOUS(src->n64.buttons, N64_BTN_L|N64_BTN_R|N64_BTN_Z|N64_BTN_DPAD_LEFT)) {
				g_n64_mapping_mode = MODE_OCARINA;
			} else if (IS_SIMULTANEOUS(src->n64.buttons, N64_BTN_L|N64_BTN_R|N64_BTN_Z|N64_BTN_DPAD_RIGHT)) {
				g_n64_mapping_mode = MODE_SSMB;
			} else if (IS_SIMULTANEOUS(src->n64.buttons, N64_BTN_L|N64_BTN_R|N64_BTN_Z|N64_BTN_C_UP)) {
				g_n64_mapping_mode = MODE_SIN_AND_PUNISHMENT;
			} else if (IS_SIMULTANEOUS(src->n64.buttons, N64_BTN_L|N64_BTN_R|N64_BTN_Z|N64_BTN_C_DOWN)) {
				g_n64_mapping_mode = MODE_OGRE_BATTLE;
			}

			if (src->n64.buttons & N64_BTN_A) { dst->buttons |= CPAD_BTN_A; }
			if (src->n64.buttons & N64_BTN_B) { dst->buttons |= CPAD_BTN_B; }
			if (src->n64.buttons & N64_BTN_START) { dst->buttons |= CPAD_BTN_PLUS; }

			if (src->n64.buttons & N64_BTN_DPAD_UP) { dst->buttons |= CPAD_BTN_DPAD_UP; }
			if (src->n64.buttons & N64_BTN_DPAD_DOWN) { dst->buttons |= CPAD_BTN_DPAD_DOWN; }
			if (src->n64.buttons & N64_BTN_DPAD_LEFT) { dst->buttons |= CPAD_BTN_DPAD_LEFT; }
			if (src->n64.buttons & N64_BTN_DPAD_RIGHT) { dst->buttons |= CPAD_BTN_DPAD_RIGHT; }

			switch (g_n64_mapping_mode)
			{
				case MODE_STANDARD:
					if (src->n64.buttons & N64_BTN_Z) { dst->buttons |= CPAD_BTN_ZL | CPAD_BTN_ZR; }
					if (src->n64.buttons & N64_BTN_R) { dst->buttons |= CPAD_BTN_TRIG_RIGHT; }
					if (src->n64.buttons & N64_BTN_L) { dst->buttons |= CPAD_BTN_TRIG_LEFT; }
					break;
				case MODE_MARIOKART64:
					if (src->n64.buttons & N64_BTN_Z) { dst->buttons |= CPAD_BTN_TRIG_LEFT; }
					if (src->n64.buttons & N64_BTN_R) { dst->buttons |= CPAD_BTN_TRIG_RIGHT; }
					if (src->n64.buttons & N64_BTN_L) { dst->buttons |= CPAD_BTN_ZL | CPAD_BTN_ZR; }
					break;
				case MODE_OCARINA:
					if (src->n64.buttons & N64_BTN_Z) { dst->buttons |= CPAD_BTN_TRIG_LEFT; }
					if (src->n64.buttons & N64_BTN_R) { dst->buttons |= CPAD_BTN_TRIG_RIGHT; }
					if (src->n64.buttons & N64_BTN_L) { dst->buttons |= CPAD_BTN_DPAD_DOWN; }
					break;
				case MODE_SSMB:
					if (src->n64.buttons & N64_BTN_L) { dst->buttons |= CPAD_BTN_DPAD_DOWN; }
					if (src->n64.buttons & N64_BTN_R) { dst->buttons |= CPAD_BTN_ZL | CPAD_BTN_ZR; }
					if (src->n64.buttons & N64_BTN_Z) { dst->buttons |= CPAD_BTN_TRIG_LEFT | CPAD_BTN_TRIG_RIGHT; }
					if (src->n64.buttons & N64_BTN_C_LEFT) { dst->buttons |= CPAD_BTN_Y; }	
					if (src->n64.buttons & N64_BTN_C_DOWN) { dst->buttons |= CPAD_BTN_X; }	
					break;
				case MODE_SIN_AND_PUNISHMENT:
					if (src->n64.buttons & N64_BTN_L) { dst->buttons |= CPAD_BTN_ZL | CPAD_BTN_ZR; }
					if (src->n64.buttons & N64_BTN_R) { dst->buttons |= CPAD_BTN_TRIG_RIGHT; }
					if (src->n64.buttons & N64_BTN_Z) { dst->buttons |= CPAD_BTN_TRIG_LEFT; }
					if (src->n64.buttons & N64_BTN_C_LEFT) { dst->buttons |= CPAD_BTN_Y; }
					if (src->n64.buttons & N64_BTN_C_RIGHT) { dst->buttons |= CPAD_BTN_X; }
					break;
				case MODE_OGRE_BATTLE:
					if (src->n64.buttons & N64_BTN_L) { dst->buttons |= CPAD_BTN_TRIG_LEFT; }
					if (src->n64.buttons & N64_BTN_R) { dst->buttons |= CPAD_BTN_TRIG_RIGHT; }
					if (src->n64.buttons & N64_BTN_Z) { dst->buttons |= CPAD_BTN_TRIG_LEFT; }
					break;
			}

			// In sin and punishment, both the left and right analog sticks on the wii classic 
			// controller are used for aim. On a N64 controller, C-left and C-right are used to 
			// move. Up down are not used and we do not want accidental button presses there to 
			// mess with aiming!
			if (g_n64_mapping_mode != MODE_SIN_AND_PUNISHMENT) {
				if (src->n64.buttons & N64_BTN_C_UP) { dst->ry = 100; }
				if (src->n64.buttons & N64_BTN_C_DOWN) { dst->ry = -100; }
				if (src->n64.buttons & N64_BTN_C_LEFT) { dst->rx = -100; }
				if (src->n64.buttons & N64_BTN_C_RIGHT) { dst->rx = 100; }
			}

			dst->lx = src->n64.x;
			dst->ly = src->n64.y;

			/*
			if (src->n64.buttons & N64_BTN_C_UP) { 	dst->ly = 100; 		} 
			if (src->n64.buttons & N64_BTN_C_DOWN) { 	dst->ly = -100; 		} 
			if (src->n64.buttons & N64_BTN_C_LEFT) { 	dst->lx = -100; 		} 
			if (src->n64.buttons & N64_BTN_C_RIGHT) { 	dst->lx = 100; 		} 
			*/
			break;
#endif
	}
}

static void pollfunc(void)
{
	performupdate = 1;
}

#define ERROR_THRESHOLD			10

#define STATE_NO_CONTROLLER		0
#define STATE_CONTROLLER_ACTIVE	1

int main(void)
{
	Gamepad *snes_gamepad = NULL;
	Gamepad *n64_gamepad = NULL;
	Gamepad *gc_gamepad	= NULL;
	Gamepad *cur_gamepad = NULL;
	unsigned char mainState = STATE_NO_CONTROLLER;
	unsigned char analog_style = ANALOG_STYLE_DEFAULT;
	gamepad_data lastReadData;
	classic_pad_data classicData;
	unsigned char current_report[6];
	int error_count = 0;
	
	
	hwInit();	


#ifdef WITH_SNES
	snes_gamepad = snesGetGamepad();
#endif

#ifdef WITH_N64 
	n64_gamepad = n64GetGamepad();
#endif

#ifdef WITH_GAMECUBE
	gc_gamepad = gamecubeGetGamepad();
#endif
	
	dataToClassic(NULL, &classicData);
	pack_classic_data(&classicData, current_report, ANALOG_STYLE_DEFAULT);

	memset(cal_data, 0, sizeof(cal_data));
	wm_init(classic_id, current_report, cal_data, pollfunc);
	sei();

	while(1)
	{
		int tmp;

		set_sleep_mode(SLEEP_MODE_EXT_STANDBY);
		sleep_enable();
		sleep_cpu();
		sleep_disable();

		while (!performupdate) { }	
		performupdate = 0;
		_delay_ms(1);

		switch(mainState)
		{
			default:
				mainState = STATE_NO_CONTROLLER;
				break;

			case STATE_NO_CONTROLLER:
#if defined(WITH_GAMECUBE) || defined(WITH_N64)
				switch (gcn64_detectController())
				{
					case CONTROLLER_IS_N64:
						cur_gamepad = n64_gamepad;
						mainState = STATE_CONTROLLER_ACTIVE;
						analog_style = ANALOG_STYLE_N64;
						break;

					case CONTROLLER_IS_GC:
						cur_gamepad = gc_gamepad;
						mainState = STATE_CONTROLLER_ACTIVE;
						analog_style = ANALOG_STYLE_GC;
						break;

					default:
					case CONTROLLER_IS_ABSENT:
					case CONTROLLER_IS_UNKNOWN:
						snes_gamepad->update();
						snes_gamepad->getReport(&lastReadData);
					break;
				}
#else
				snes_gamepad->update();
				snes_gamepad->getReport(&lastReadData);
#endif
				break;
			
			case STATE_CONTROLLER_ACTIVE:
				if (cur_gamepad->update()) {
					error_count++;
					if (error_count > 10)
						mainState = STATE_NO_CONTROLLER;
					break;
				}
				error_count = 0;
				cur_gamepad->getReport(&lastReadData);
				break;
		}
	
		dataToClassic(&lastReadData, &classicData);		
		pack_classic_data(&classicData, current_report, analog_style);
		wm_newaction(current_report);
	}

	return 0;
}
